import os

import lumeo.scripts.gateway.hostutils as hostutils
import lumeo.scripts.gateway.display as display

def install_update_gateway_updater():
    display.output_message("Installing/updating gateway updater...", status='info')
    # Uninstall lumeo-container-update cron script if installed
    if os.path.exists("/etc/cron.hourly/lumeo-container-update"):
        hostutils.run_command("rm -f /etc/cron.hourly/lumeo-container-update", sudo=True)
        display.output_message("Removed old lumeo-container-update cron script", status='info')

    # Uninstall lumeo-wgm-update cron script if installed
    if os.path.exists("/etc/cron.hourly/lumeo-wgm-update-cron"):
        hostutils.run_command("rm -f /etc/cron.hourly/lumeo-wgm-update-cron", sudo=True)
        display.output_message("Removed old lumeo-wgm-update-cron cron script", status='info')

    # Install the new, unified lumeo-update-cron.sh script
    new_update_script_name = "lumeo-gateway-update" # TODO: rename to lumeo-update-cron when we start publishing to the web
    if not os.path.exists(f"/etc/cron.hourly/{new_update_script_name}"):    
        script_path = os.path.join(os.path.dirname(__file__), f"{new_update_script_name}.sh")
        hostutils.run_command(f"install -m u=rwx,g=rx,o=rx {script_path} /etc/cron.hourly/{new_update_script_name}", sudo=True)


import shutil
import platform
import json
import os
import docker
import requests

import lumeo.scripts.gateway.display as display
import lumeo.scripts.gateway.hostutils as hostutils
import lumeo.scripts.gateway.nvidia as nvidia

DOCKER_CONFIG_FILE = "/etc/docker/daemon.json"
DOCKER_CLIENT = docker.from_env()
LUMEO_WATCHTOWER_VERSION = "2"


################################################################################
# Docker Installation & Configuration
################################################################################

def install_docker_and_nvidia_toolkit():
    """Install Docker and NVIDIA toolkit if necessary."""
    if shutil.which("docker") is None:
        hostutils.apt_install("docker.io containerd", update_first=True)
    
    hostutils.run_command("systemctl start docker", sudo=True)
    
    arch_type = platform.machine()
    l4t_ver = nvidia.l4t_version() if arch_type == "aarch64" else None
    
    if arch_type == "x86_64" or (arch_type == "aarch64" and l4t_ver and l4t_ver >= "36"):
        if not hostutils.run_command("dpkg-query -W -f='${Status}' nvidia-container-toolkit", check=False):
            hostutils.apt_install("curl gnupg", update_first=True)
            hostutils.run_command("curl -s -L https://nvidia.github.io/libnvidia-container/gpgkey | sudo gpg --dearmor -o /usr/share/keyrings/nvidia-container-toolkit-keyring.gpg --yes", shell=True)
            hostutils.run_command("curl -s -L https://nvidia.github.io/libnvidia-container/stable/deb/nvidia-container-toolkit.list | sed 's#deb https://#deb [signed-by=/usr/share/keyrings/nvidia-container-toolkit-keyring.gpg] https://#g' | sudo tee /etc/apt/sources.list.d/nvidia-container-toolkit.list > /dev/null", shell=True)
            hostutils.apt_install("nvidia-container-toolkit", update_first=True)
            hostutils.run_command("nvidia-ctk runtime configure --runtime=docker", sudo=True)
            hostutils.run_command("systemctl restart docker", sudo=True)
    
    elif arch_type == "aarch64" and l4t_ver and l4t_ver < "36":
        if not hostutils.run_command("dpkg -l nvidia-docker2", check=False):
            hostutils.apt_install("nvidia-docker2", update_first=True)
            hostutils.run_command("systemctl restart docker", sudo=True)


def check_set_docker_data_dir():
    
    move_docker_data_dir = hostutils.prompt_yes_no("Use external drive to store the Docker data directory? (for docker images & volumes)? (Recommended if your root partition is smaller than 32 GB). Check instructions how to mount the external drive in https://docs.lumeo.com/. Proceed?", "n")
    
    if move_docker_data_dir:
        
        external_folder = hostutils.prompt_input("Enter the path to the external drive where Docker data directory will be moved. (e.g. /mnt/nvme1n1p1)")
        
        original_folder = hostutils.run_command("docker info -f '{{ .DockerRootDir}}'", sudo=True)

        if not os.path.isdir(external_folder):
            display.output_message(f"Error: Directory '{external_folder}' does not exist. Please make sure the external drive is correctly mounted.", status='error')
            exit(1)

        hostutils.run_command("systemctl stop docker docker.socket containerd", sudo=True)

        if os.path.isdir(original_folder):
            available_space = int(hostutils.run_command(f"df -P {external_folder} | awk 'END{{print $4}}'", sudo=True))
            original_folder_size = int(hostutils.run_command(f"du -s {original_folder} | awk '{{print $1}}'", sudo=True))

            if original_folder_size > available_space:
                display.output_message(f"Error: There's not enough space on '{external_folder}' to move the contents of '{original_folder}'", status='error')
                hostutils.run_command("systemctl start docker", sudo=True)
                exit(1)
            else:
                display.output_message(f"Moving the contents of '{original_folder}' to '{external_folder}/docker'. Please wait a moment...")
                hostutils.run_command(f"cp -rp {original_folder} {external_folder}", sudo=True)
        else:
            display.output_message(f"Error: Directory '{original_folder}' does not exist. Please make sure Docker is correctly installed.", status='error')
            hostutils.run_command("systemctl start docker", sudo=True)
            exit(1)

        hostutils.run_command(f"mv {original_folder} {original_folder}.old", sudo=True)
        update_docker_data_directory(f"{external_folder}/docker")
        display.output_message(f"Successfully updated docker daemon data folder to '{external_folder}/docker'. Backups can be found in '{original_folder}.old' directory")
        hostutils.run_command("systemctl start docker", sudo=True)
        
        
def update_docker_data_directory(docker_data_folder):
    display.output_message("[lumeo-container-manager] Updating docker daemon data-root")
    docker_config_file = "/etc/docker/daemon.json"

    with open(docker_config_file, "r") as json_file:
        docker_config = json.load(json_file)
        docker_config["data-root"] = docker_data_folder

    with open(docker_config_file, "w") as json_file:
        json.dump(docker_config, json_file)

        
# Sets up logging configuration for docker daemon.
# This applies only to newly created containers. 
# During lumeod update, the container gets re-created.
def set_docker_logs_configuration():
    # Check if the file exists
    if os.path.isfile(DOCKER_CONFIG_FILE):
        # If the file exists, open it and load the JSON data
        with open(DOCKER_CONFIG_FILE, 'r') as f:
            config = json.load(f)
    else:
        # If the file doesn't exist, create an empty dictionary to use as the config
        config = {}

    # If we had already set this, exit.
    if "log-opts" in config:
        display.output_message(f"[set_docker_logs_configuration] 'log-opts' are already set to {config['log-opts']}, skipping ...", status='debug')
        return

    # Add the "log-opts" entry to the config
    config["log-opts"] = {
        "max-size": "10m",
        "max-file": "1"
    }

    # Write the config back to the file
    with open(DOCKER_CONFIG_FILE, 'w') as f:
        json.dump(config, f, indent=4)

    display.output_message("[set_docker_logs_configuration] Docker config updated, restarting docker daemon ...", status='info')
    hostutils.run_command("systemctl restart docker", sudo=True)
    display.output_message("[set_docker_logs_configuration] Done.", status='info')

    
# This fixes "Failed to initialize NVML: Unknown Error".
# The error happens when someone (usually apt) does `systemctl daemon-reload`.
# GPUs disappear from the container but still available on the host and lumeod container restart helps.
# The official workaround is to switch back to the old `cgroups` implementation: `cgroupfs`.
# For more information see:
# - https://github.com/NVIDIA/nvidia-docker/issues/1730
# - https://github.com/NVIDIA/nvidia-docker/issues/1671
# TODO: switch this back when NVIDIA releases the correct way to deal with this.
def fix_nvidia_nvml_init_issue():
    # TODO: Don't do this on OpenVINO when we have more precise platform specification.
    arch_type = platform.machine()
    if arch_type != "x86_64":
        display.output_message("[fix_nvidia_nvml_init_issue] Only x86 is affected, skipping ...", status='debug')
        return

    # Check if the file exists
    if os.path.isfile(DOCKER_CONFIG_FILE):
        # If the file exists, open it and load the JSON data
        with open(DOCKER_CONFIG_FILE, 'r') as f:
            config = json.load(f)
    else:
        # If the file doesn't exist, create an empty dictionary to use as the config
        config = {}

    # If we had already set this, exit.
    # If someone else had set 'exec-opts', let's exit as well not to mess things up.
    if "exec-opts" in config:
        display.output_message(f"[fix_nvidia_nvml_init_issue] 'exec-opts' are already set to {config['exec-opts']}, skipping ...", status='debug')
        return

    # Add the "exec-opts" entry to the config
    config["exec-opts"] = ["native.cgroupdriver=cgroupfs"]

    # Write the config back to the file
    with open(DOCKER_CONFIG_FILE, 'w') as f:
        json.dump(config, f, indent=4)

    display.output_message("[fix_nvidia_nvml_init_issue] Docker config updated, restarting docker daemon ...", status='info')
    hostutils.run_command("systemctl restart docker", sudo=True)
    display.output_message("[fix_nvidia_nvml_init_issue] Done.", status='info')
    
################################################################################
# Docker container operations
################################################################################

def check_container_exists(container_name):
    all_containers = DOCKER_CLIENT.containers.list(all=True)
    for container in all_containers:
        if container_name == container.name:
            return True
    return False

def stop_and_remove_docker_containers(container_prefix="lumeo/gateway", silent=False):
    """Stop and remove Docker containers with image name starting with the specified prefix."""    
    display.output_message(f"Checking for running Docker containers with image name starting with '{container_prefix}'. Will stop and remove.")    
    docker_containers = hostutils.run_command(f"docker ps -a --format '{{{{.Image}}}} {{{{.ID}}}}' | grep '^{container_prefix}' | awk '{{print $2}}'", sudo=True)
    
    if docker_containers:
        if not silent:
            display.output_message("Stopping and removing Docker containers...")
        for container in docker_containers.split():
            hostutils.run_command(f"docker stop {container}", sudo=True)
            hostutils.run_command(f"docker rm -v {container}", sudo=True)
    else:
        display.output_message(f"No Docker containers with image name starting with '{container_prefix}' found.")
        
        
def docker_download_image(image_name):
    display.output_message(f"Downloading the container version {image_name}...", "info")
    display.get_progress(reset=True)
    display.add_progress_task("docker_download_image", f"Downloading {image_name}")    
    
    try:
        for line in DOCKER_CLIENT.api.pull(image_name, stream=True, decode=True):
            if 'progress' in line:
                progress_str = line['progress'].strip()
                if progress_str.endswith('%'):
                    percentage = float(progress_str[:-1])
                    display.update_progress_task('docker_download_image', completed=percentage)
            elif 'status' in line and line['status'] == 'Download complete':
                display.update_progress_task('docker_download_image', completed=100)
    except Exception as e:
        display.output_message(f"Error downloading container: {str(e)}", status='error')
        raise
    finally:
        display.update_progress_task('docker_download_image', completed=100)
    
    display.stop_progress()
    display.output_message(f"Container version {image_name} downloaded.", "info")


################################################################################
# Watchtower
################################################################################

def start_watchtower():
    watchtower_name = "lumeo-watchtower"
    if not DOCKER_CLIENT.containers.list(all=True, filters={"name": watchtower_name}):
        watchtower_labels = {
            "lumeo.container_version": LUMEO_WATCHTOWER_VERSION,
            "lumeo.container_type": "containrrr/watchtower",
        }
        DOCKER_CLIENT.containers.run(
            "containrrr/watchtower:1.4.0",
            command=["--http-api-update", "--http-api-token", "lumeo-container-update-watchtower-token",
                     "--label-enable", "--include-stopped", "--cleanup"],
            volumes=["/var/run/docker.sock:/var/run/docker.sock"],
            restart_policy={"MaximumRetryCount": 0, "Name": "always"},
            name=watchtower_name,
            labels=watchtower_labels,
            ports={'8080/tcp': ('127.0.0.1', 46655)},
            detach=True,
        )


# Re-creates watchtower if the current version is other than LUMEO_WATCHTOWER_VERSION
def migrate_watchtower():
    try:
        # Get the watchtower container
        watchtower = DOCKER_CLIENT.containers.get("lumeo-watchtower")

        # Get the current version of the watchtower container
        current_version = watchtower.attrs["Config"]["Labels"]["lumeo.container_version"]

        # Migrate 1 to 2.
        if current_version == "1":
            # Remove the existing watchtower container
            watchtower.remove(force=True)

            # Start a new watchtower container with the updated version
            start_watchtower()

            display.output_message("Watchtower migrated to the latest version.", "success")
        else:
            display.output_message("Watchtower is already up to date.", "info")
    except docker.errors.NotFound:
        display.output_message("Watchtower container not found.", "info")
    except (KeyError, ValueError):
        display.output_message("Error parsing watchtower version.", "error")
    except Exception as e:
        display.output_message(f"Error migrating watchtower: {str(e)}", "error")


def watchtower_trigger_update():
    display.output_message("Checking for updates in Lumeo containers images")
    # Ask watchtower to check for container image updates
    token = "lumeo-container-update-watchtower-token"
    final_url = "http://localhost:46655/v1/update"
    headers_api = {
        "Authorization": "Bearer " + token
    }
    try:
        requests.get(url=final_url, headers=headers_api)
    except Exception as err:
        display.output_message("[lumeo-container-update] Exception: {}".format(err), "error")




import base64
import gzip
import os
import json
import logging

from rich import print
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from rich.progress import Progress, SpinnerColumn, TimeElapsedColumn
from rich.prompt import Prompt
from rich.table import Table
from rich.logging import RichHandler


# handler.setFormatter(colorlog.ColoredFormatter(
#     "%(log_color)s%(asctime)s %(levelname)-8s %(message)s",
#     datefmt='%Y-%m-%d %H:%M:%S',
#     log_colors={
#         'DEBUG': 'cyan',
#         'INFO': 'green',
#         'WARNING': 'yellow',
#         'ERROR': 'red,bg_white',
#         'CRITICAL': 'red,bg_white',
#     },
# ))
FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET", format=FORMAT, datefmt="[%X]", 
    handlers=[RichHandler(markup=True, show_path=False)]
)

logger = logging.getLogger("lumeo")
logger.setLevel(logging.INFO)

urllib3_logger = logging.getLogger("urllib3")
urllib3_logger.setLevel(logging.WARNING)
docker_logger = logging.getLogger("docker")
docker_logger.setLevel(logging.WARNING)

console = Console()
progress = None
progress_tasks = {}
    
OUTPUT_FORMAT = 'text'
    
def set_output_format(format):
    global OUTPUT_FORMAT
    OUTPUT_FORMAT = 'json' if format in ['json', True] else 'text'
    
def lumeo_logo():
    encoded_data = "H4sIACZdkGMCA8WUMQ7AIAhFd07hURkcHBicPKAnaUxqqy1qVbTNnxr8L5APSt2fd1pEkBmaZYqgxRgG6Z3dRQySg2KiZjFF7iArTRYOMbD0EvPAVihlf+JmSkz11YSd8W6PkqkYIFRCQl2EEHd+t/hpY8FpIkVaPc+TEfW/QScrJurLVvxwAtRGJEFzeaW7S3gCyky5e5H8gFelnZhYZzMAB0lpiLmGBwAA"
    decoded_data = base64.b64decode(encoded_data)
    decompressed_data = gzip.decompress(decoded_data)
    return decompressed_data.decode('utf-8')
    
def print_header(text, title=None, subtitle=None):
    if title:
        console.print(Panel(text, title=title, subtitle=subtitle, style="cyan"))
    else:
        console.print(Panel(text, style="cyan"))
    
def print_text(text):
    console.print(text)

def output_message(message, status='info', title=''):
    if OUTPUT_FORMAT == 'json':
        output = {'status': status, 'message': message}
        print_text(json.dumps(output))
    else:
        if status == 'debug':
            logger.debug(message)
        elif status == 'warning':
            logger.warning(message)
        elif status == 'error':
            logger.error(message)
        else:
            logger.info(message)
        #print_text(title)
        #print_text(message)        

def output_data(headers, rows, title, expand=True):
    if OUTPUT_FORMAT == 'json':
        # Convert headers and rows into a list of dictionaries
        output = {'status': 'success', 'data': []}
        headers_fixed = [header.replace(" ", "_").lower() for header in headers]
        for row in rows:
            output['data'].append(dict(zip(headers_fixed, row)))
        print(json.dumps(output))
    else:
        table = generate_table(headers, rows, title, expand)
        console.print(table)
        
def get_progress(reset=False):
    global progress, progress_tasks
    if reset or not progress:
        progress = Progress()
        progress_tasks = {}
        progress.start()
    return progress

def stop_progress():
    global progress, progress_tasks
    progress.stop()
    progress_tasks = {}

def add_progress_task(task_key, description, total=100):
    progress_tasks[task_key] = (get_progress().add_task(description, total=total), total)
    return progress_tasks[task_key]

def update_progress_task(task_key, advance=1, completed=None):
    if completed and progress_tasks[task_key][1] is None:
        get_progress().update(progress_tasks[task_key][0], total=completed)
    get_progress().update(progress_tasks[task_key][0], advance=advance, completed=completed)
    
        
def generate_table(headers, rows, title, expand=True):
    table = Table(title=title, expand=expand, style="cyan")
    for header in headers:
        table.add_column(header, style="cyan", no_wrap=False)
    for row in rows:
        table.add_row(*row)
    return table

def prompt_yes_no(msg, default_value):
    """Prompt for yes/no input with a default value."""
    while True:
        if os.getenv('NO_PROMPT', False):
            console.print(f"{msg} (using default: {default_value})")
            return default_value.lower() == 'y'

        response = prompt_input(msg, default=default_value.lower(), choices=['y', 'n'])
        if response in ['y', 'n']:
            return response == 'y'
        elif response == '':
            return default_value.lower() == 'y'
        else:
            print("Please enter 'y' for Yes, 'n' for No, or press Enter for the default value.")            


def prompt_input(msg, default=None, choices=None):
    """Prompt for input with a default value."""
    if os.getenv('NO_PROMPT', False):
        console.print(f"{msg} (using default: {default})")
        return default
    return Prompt.ask(f"[bold]{msg}[/bold]", default=default, choices=choices)



import os
import shutil
import subprocess
import sys
import time
from rich.progress import Progress

from .display import prompt_yes_no, output_message

# Common host utilities
def check_os():
    """Check if the OS is Linux."""
    if os.name != "posix" or os.uname().sysname != "Linux":
        output_message("[check_os] This script must be run on a Linux system.", status='error')
        sys.exit(1)

def run_command(command, check=True, shell=True, sudo=False, error_message=None, useOsRun=False, silent=False):
    """Run a command and return its output."""
    try:
        command_name = command.split()[0]  # Get the name of the command
        if shutil.which(command_name) is None:  # Check if the command is available
            if not silent:
                output_message(f"[run_command] Command '{command}' not found.", status='error')
            return None                
        
        if sudo:
            command = f"sudo {command}"
        
        if not shell:
            command = command.split()
            
        if useOsRun:
            result = os.system(command)
            return result
        else:
            result = subprocess.run(command, check=check, shell=shell, text=True,
                                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        output_message(f"[run_command] Error executing command: {e}", status='error')
        output_message(f"[run_command] Command output: {e.output}", status='error')
        output_message(f"[run_command] Command error: {e.stderr}", status='error')
        if error_message:
            output_message(f"[run_command] {error_message}", status='error')
        sys.exit(1)

def check_lock(command):
    """Check for apt/dpkg locks and wait if necessary."""
    max_attempts = 300
    with Progress() as progress:
        task = progress.add_task("[cyan]Waiting for locks to be released...", total=max_attempts)
        for i in range(max_attempts):
            if not run_command("sudo fuser /var/lib/dpkg/lock /var/lib/apt/lists/lock /var/cache/apt/archives/lock", check=False):
                break
            progress.update(task, advance=1)
            time.sleep(1)
        else:
            output_message("[check_lock] Unable to acquire lock after 5 minutes. Aborting installation.", status='error')
            sys.exit(1)
    
    run_command(command, shell=True, sudo=True)

def apt_install(packages, update_first=False):
    """Install packages using apt, optionally updating first."""
    if update_first:
        check_lock("apt update")
    check_lock(f"apt install -y {packages}")


# Install host packages
def setup_host_common():
    apt_install("curl jq python3 lshw util-linux smartmontools", update_first=True)

        

def prompt_disable_x_server():
    """Prompt to disable X server and return whether a reboot is required."""
    if prompt_yes_no("[Optional] Disable GUI (X server) to free up memory, allowing you to run more pipelines? (reboot required)", "n"):
        run_command("systemctl enable multi-user.target", sudo=True)
        run_command("systemctl set-default multi-user.target", sudo=True)
        return "In order to turn off GUI (X server), a reboot is required"
    return None


def toggle_ssh(enable=False):
    """Enable or disable SSH."""
    action = "enable" if enable else "disable"
    if prompt_yes_no(f"Do you want to {action} SSH?", "n"):
        output_message(f"[toggle_ssh] {'Enabling' if enable else 'Disabling'} SSH...", status='info')
        run_command(f"systemctl {action} ssh", sudo=True)
        output_message(f"[toggle_ssh] SSH has been {action}d.", status='info')
    else:
        output_message("[toggle_ssh] Operation cancelled.", status='info')


def toggle_user_lock(username, lock=True):
    """Lock or unlock the user account."""    
    action = "lock" if lock else "unlock"
    if prompt_yes_no(f"Do you want to {action} the password for user {username}?", "n"):
        output_message(f"[toggle_user_lock] {'Locking' if lock else 'Unlocking'} password for user {username}...", status='info')
        run_command(f"passwd -{action[0]} {username}", sudo=True)
        output_message(f"[toggle_user_lock] User {username} has been {action}ed.", status='info')
    else:
        output_message("[toggle_user_lock] Operation cancelled.", status='info')
        

def check_disk_space():
    docker_data_folder = run_command("docker info -f '{{ .DockerRootDir}}'", sudo=True)
    if os.path.isdir(docker_data_folder):
        arch_type = os.uname().machine
        required_size_KB = 8388608 if arch_type == "aarch64" else 20971520

        available_space_KB = int(run_command(f"df -P {docker_data_folder} | awk 'END{{print $4}}'", sudo=True))

        if available_space_KB < required_size_KB:
            required_size_GB = required_size_KB / (1024 ** 2)
            available_space_GB = available_space_KB / (1024 ** 2)

            output_message("WARNING: There is not enough free space on the disk where Docker stores its data."
                           f"You have {available_space_GB:.2f} GB available, but at least {required_size_GB:.2f} GB is required for fresh lumeod installations."
                           "If the container image has not been pulled yet, or lumeod is not installed, you might run out of disk space during installation.", 
                           status='warning')

            if prompt_yes_no("Do you want to continue despite the low disk space?", "n"):
                output_message("Proceeding despite the low disk space warning.", status='warning')
            else:
                output_message("Exiting due to insufficient disk space.", status='error')
                exit(1)



from .metadata import (
    num_milliseconds_seen,
    get_timestamp,
    eval_trigger_condition,
    merge_dicts,
    flatten_dict,
    filter_dict,
    object_matches_type,
    object_matches_types,
    object_has_attributes,
)


__all__ = [
    'num_milliseconds_seen', 'get_timestamp', 'eval_trigger_condition',
    'merge_dicts', 'flatten_dict', 'filter_dict', 'object_matches_type',
    'object_matches_types', 'object_has_attributes'
]



import argparse
import os
import sys

from lumeo.utils import check_for_update
import lumeo.scripts.gateway.display as display
import lumeo.scripts.gateway.lumeod as lumeod
import lumeo.scripts.gateway.wgm as wgm
import lumeo.scripts.gateway.schedule as schedule
import lumeo.scripts.gateway.docker as dockerutils


def exit_manager():
    sys.exit(0)
    
COMMANDS = {
    "list": ("List the currently installed Lumeo gateway containers", lumeod.list_containers, True),
    "install": ("Install a new Lumeo gateway container", lumeod.install_container, True),
    "download": ("Download the latest Lumeo gateway container image (OEM use only)", lumeod.download_container, True),
    "stop": ("Kills a running Lumeo gateway container", lumeod.stop_container, True),
    "restart": ("Restart a Lumeo gateway container & update to the latest version", lumeod.upgrade_and_restart_container, True),
    "update": ("Update all Lumeo gateway containers to the latest image", lumeod.update_containers, True),
    "remove": ("Remove a Lumeo gateway container", lumeod.remove_container, True),
    "logs": ("Print the logs of a Lumeo gateway container", lumeod.logs, True),
    "shell": ("Open an interactive shell in a running Lumeo gateway container", lumeod.shell, True),

    "schedule": ("Schedule container update in a certain date range", schedule.edit_update_schedule, True),
    
    "install-wgm": ("Install Web Gateway Manager", wgm.install_wgm, True),
    "update-wgm": ("Update Web Gateway Manager", wgm.update_wgm, True),
    "reset-wgm": ("Reset Web Gateway Manager", wgm.reset_wgm, True),
    "remove-wgm": ("Remove Web Gateway Manager", wgm.remove_wgm, True),
    
    "fix_nvidia_driver_issue": ("Fix the issue with the NVIDIA driver", dockerutils.fix_nvidia_nvml_init_issue, False),
    "set_docker_logs_configuration": ("Set the Docker logs configuration", dockerutils.set_docker_logs_configuration, False),
    "write_hardware_information": ("Write the hardware information to a file", lumeod.write_hardware_information_to_container, False),    
    "start_watchtower": ("Start Watchtower", dockerutils.start_watchtower, False),
    "migrate_watchtower": ("Migrate Watchtower", dockerutils.migrate_watchtower, False),
    
    "exit": ("Quits this script", exit_manager, True),
}

def print_menu():
    update_info = check_for_update()
    _, wgm_status_str = wgm.get_wgm_status()
    lumeo_containers_all = lumeod.get_lumeo_containers(show_all=True)
    lumeo_containers_running = lumeod.get_lumeo_containers(show_all=False)
    subtitle = f"Status | Lumeo Gateways: [bold green]{len(lumeo_containers_running)} running[/bold green], [bold]{len(lumeo_containers_all)} total[/bold] | {wgm_status_str}"
    print('')
    display.print_header(get_commands(), f"[bold red]BETA[/bold red] [bold cyan]Lumeo Gateway Manager v{update_info[2]}[/bold cyan]", subtitle=subtitle)

def get_commands():
    headers = ["Command", "Description"]
    rows = [
        [command_name, description]
        for command_name, (description, _, show_in_menu) in COMMANDS.items()
        if show_in_menu
    ]
    return display.generate_table(headers, rows, "Commands", expand=False)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("command", choices=COMMANDS.keys(), nargs="?")
    parser.add_argument("--container-name")
    parser.add_argument("--force-update", action="store_true", help="Force update without checking schedule")
    parser.add_argument("--no-prompt", action="store_true", help="Script mode. Uses defaults, does not prompt for input", default=False)
    parser.add_argument("--app-id", help="Lumeo workspace / app id to link a gateway to")
    parser.add_argument("--api-key", help="Lumeo workspace/app API key")
    parser.add_argument("--gateway-name", help="Gateway name to be used for a new `install` with `--no-prompt` options.")
    parser.add_argument("--json", action="store_true", help="Output in JSON format")

    args = parser.parse_args()
    
    if args.no_prompt:
        os.environ['NO_PROMPT'] = 'true'
    if args.app_id:
        os.environ['LUMEO_APP_ID'] = args.app_id
    if args.api_key:
        os.environ['LUMEO_API_KEY'] = args.api_key
    if args.gateway_name:
        os.environ['LUMEO_GATEWAY_NAME'] = args.gateway_name
    if args.json:  
        # Set output format
        display.set_output_format(args.json)
        os.environ['NO_PROMPT'] = 'true'

    if not args.command:
        containers_list = lumeod.get_lumeo_containers()

        if not containers_list:
            if os.getenv('NO_PROVISION', False):
                # pull the latest version of lumeo image, but not attempt to install it
                lumeod.download_container()
                sys.exit(0)
            else:
                # start a new container installation if no containers are found
                lumeod.install_container()

        print_menu()

        # if Lumeo containers are found, we list them
        lumeod.list_containers()

        while True:
            try:
                user_option = display.prompt_input("Enter the command")
                if user_option in COMMANDS:
                    COMMANDS[user_option][1]()
                else:
                    print("Error: Command is not recognized! Please select a valid command\n")
                    print_menu()
            except KeyboardInterrupt:
                sys.exit(0)
    else:
        # Perform a single operation and exit
        if args.container_name and args.command in ["stop", "restart", "remove", "logs", "shell", "wgm",
                                                    "write_hardware_information"]:
            COMMANDS[args.command][1](args.container_name)
        elif args.force_update and args.command == "update":
            COMMANDS[args.command][1](args.force_update)
        else:
            COMMANDS[args.command][1]()
            
            
if __name__ == "__main__":
    main()


#!/usr/bin/env python3
import os
import sys
import petname
import toml
import docker
import json
import shutil

import lumeo.scripts.gateway.display as display
import lumeo.scripts.gateway.hostutils as hostutils
import lumeo.scripts.gateway.nvidia as nvidia
import lumeo.scripts.gateway.docker as dockerutils
import lumeo.scripts.gateway.common as common
import lumeo.scripts.gateway.updater as updater
import lumeo.scripts.gateway.schedule as schedule

# Container versions:
# 1. Initial version
# 2. `--gpus all` on x86_64` vs `--runtime nvidia` on `aarch64`
# 3. '--privileged' mode is now required
LUMEO_CONTAINER_VERSION = "3"

# Get the information of current OS platform
os_platform_info = os.uname().machine

def get_lumeod_image():
    image = os.getenv("LUMEO_IMAGE")
    if not image:
        if os_platform_info == "aarch64":
            l4t_ver = nvidia.l4t_version()
            if l4t_ver and l4t_ver < "36":
                return "lumeo/gateway-nvidia-jetson"
            elif l4t_ver and l4t_ver >= "36":
                return "lumeo/gateway-nvidia-jetson-jetpack6"
            else:
                sys.exit("Error: Could not determine the Jetpack L4T version")
        elif os_platform_info == "x86_64":
            return "lumeo/gateway-nvidia-dgpu"
        else:
            sys.exit("Unsupported platform")
    return image


def list_containers(status="all"):
    if status == "running":
        show_all = False
    else:
        show_all = True

    containers_list = get_lumeo_containers(show_all=show_all)

    containers_names = []

    if containers_list:
        container_full_info = []
        for (idx, container) in enumerate(containers_list):
            containers_names.append(container.name)
            container_full_info.append(
                [
                    str(idx + 1),
                    container.name,
                    container.status,
                    container.attrs["Config"]["Labels"]["lumeo.gateway_id"],
                    container.attrs["Config"]["Labels"]["lumeo.application_id"],
                    container.labels["lumeo.container_version"],
                    ",".join(container.image.tags)
                ]
            )

        headers = ["Index", "Name", "Status", "Gateway ID", "Application ID", "Version", "Image"]
        display.output_data(headers=headers, rows=container_full_info, title="\nList of Lumeo gateway containers installed:")
    else:
        display.output_message("No Lumeo gateway containers found", "info", title="\nList of Lumeo gateway containers installed:")
    return containers_names


def choose_container_with_prompt(status, action_description):
    containers_list = list_containers(status)
    if not containers_list:
        return None

    index_or_container_name = display.prompt_input(f"Insert the index or the name of the container to {action_description}:")

    try:
        index = int(index_or_container_name)
        if 1 <= index <= len(containers_list):
            return containers_list[index - 1]
        else:
            print("Invalid index")
            return None
    except ValueError:
        # If we couldn't parse input as a number, use the original string as a container name.
        container_name = index_or_container_name
        if container_name in containers_list:
            return container_name
        else:
            print("Container not found")
            return None




def device_requests():
    if os_platform_info == "aarch64":
        if nvidia.l4t_version() < "35.1":
            # "--gpus all" was required before JetPack 5.0.
            return [docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])]
        else:
            return None
    elif os_platform_info == "x86_64":
        return [docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])]
    else:
        raise RuntimeError("Unsupported platform: {}".format(os_platform_info))


def get_lumeo_containers(show_all=True, version=None):
    labels = []
    if version:
        labels.append("lumeo.container_version={}".format(version))

    containers_list = dockerutils.DOCKER_CLIENT.containers.list(
        all=show_all,
        filters={"label": labels},
    )

    return [container for container in containers_list 
            if "lumeo.gateway_id" in container.attrs["Config"]["Labels"]]


def download_container():
    # The only thing we do for OEM installs is to pull the latest
    # version of lumeo image. The rest will be done when the user
    # links a container to their account.
    image_name = get_lumeod_image()
    dockerutils.docker_download_image(image_name)
    return

def install_container():
    
    display.output_message("Installing Lumeo container...", "info")
        
    no_prompt = os.getenv("NO_PROMPT")
    lumeo_app_id = os.getenv("LUMEO_APP_ID")
    lumeo_api_key = os.getenv("LUMEO_API_KEY")
    lumeo_gateway_name = os.getenv("LUMEO_GATEWAY_NAME")
    lumeo_environment = os.getenv("LUMEO_ENVIRONMENT")
    lumeo_container_name_prefix = "lumeo-container-"
    auto_provision = lumeo_app_id and lumeo_api_key
    
    if no_prompt and not auto_provision:
        display.output_message("Error: `--no-prompt` requires `--app-id` and `--api-key`", "error")
        return
    
    common.install_common_dependencies()
    
    updater.install_update_gateway_updater()
                    
    # Pull the latest version of lumeo image    
    lumeo_image = get_lumeod_image()
    dockerutils.docker_download_image(lumeo_image)

    container_name = lumeo_gateway_name.replace(" ", "-").replace(":", "") if lumeo_gateway_name else lumeo_container_name_prefix + petname.Generate()

    user_container_name = display.prompt_input(f"Insert container name (or just press Enter for use default '{container_name}'):", default=container_name)
    if user_container_name:
        container_name = user_container_name

    if not dockerutils.check_container_exists(container_name):
        container_env = []
        if lumeo_environment:
            container_env.append(f"--env LUMEO_ENVIRONMENT=\'{lumeo_environment}\'")
        if not auto_provision:
            display.print_text("[bold]Please login into your Lumeo account[/bold]")
            container_env.append(f"--env CONTAINER_MODEL=\'Container {lumeo_image}\'")
        else:
            display.output_message(f"Performing the provision of Lumeo container '{container_name}' using the provided LUMEO_APP_ID & LUMEO_API_KEY):", "info")
            container_env.append(f"--env CONTAINER_MODEL=\'Container {lumeo_image}\'")
            container_env.append(f"--env LUMEO_API_KEY=\'{lumeo_api_key}\'")
            container_env.append(f"--env LUMEO_APP_ID=\'{lumeo_app_id}\'")

        # Got to perform this with system call as docker-py does not support interactive container execution
        if auto_provision:
            docker_cmd = "docker run -v {container}:/var/lib/lumeo --name {container} {env_str} --hostname {container} {image} provision"
        else:
            docker_cmd = "docker run -v {container}:/var/lib/lumeo -it --name {container} {env_str} --hostname {container} {image} provision"
        docker_cmd = docker_cmd.format(container=container_name, env_str=" ".join(container_env), image=lumeo_image)

        if hostutils.run_command(docker_cmd, sudo=True, useOsRun=True) == 0:
            dockerutils.DOCKER_CLIENT.api.remove_container(container_name, force=True)

            # At this point the volume has already been created, so we can write hardware info to it.
            write_hardware_information_to_container(container_name)

            # Read provision info from volume
            volume = dockerutils.DOCKER_CLIENT.volumes.get(container_name)
            mountpoint = volume.attrs.get("Mountpoint")

            toml_file = open(os.path.join(mountpoint, "lumeod.toml"), "r")
            lumeod_config = toml.load(toml_file)

            container_labels = {
                "com.centurylinklabs.watchtower.enable": "true",
                "lumeo.container_version": LUMEO_CONTAINER_VERSION,
                "lumeo.container_type": get_lumeod_image(),
                "lumeo.organization_id": lumeod_config["organization_id"],
                "lumeo.application_id": lumeod_config["application_id"],
                "lumeo.gateway_id": lumeod_config["gateway_id"],
            }

            # Temporary workaround until we pass the port mapping to provision
            if not dockerutils.DOCKER_CLIENT.containers.list(all=True,
                                                 filters={"label": "lumeo.primary_container"}
                                                 ):
                container_labels["lumeo.primary_container"] = "true"
                network_mode = "host"
                port_mapping = None
            else:
                network_mode = None
                port_mapping = {"8060/tcp": None}

            display.output_message("Starting container...", "info")
            
            jetson_model_name = nvidia.get_jetson_model_name()
            environment = {"JETSON_MODEL_NAME": jetson_model_name} if jetson_model_name else None

            if dockerutils.DOCKER_CLIENT.containers.run(
                    lumeo_image,
                    volumes=[container_name + ":/var/lib/lumeo"],
                    restart_policy={"MaximumRetryCount": 0, "Name": "always"},
                    name=container_name,
                    hostname=container_name,
                    network_mode=network_mode,
                    ports=port_mapping,
                    devices=nvidia.get_devices_for_container(),
                    device_requests=device_requests(),
                    runtime=nvidia.get_nvidia_docker_runtime(),
                    labels=container_labels,
                    detach=True,
                    environment=environment,
                    privileged=True,
                    ulimits=[docker.types.Ulimit(name="core", soft=0, hard=0)],
                    # Directly set ulimits to disable core dumps
            ):
                display.output_message("Container started!", "success")
        else:
            dockerutils.DOCKER_CLIENT.api.remove_container(container_name, force=True)
            display.output_message("Error: Provision on Lumeo failed", "error")
    else:
        display.output_message("Error: A container with name '{}' already exists".format(container_name), "error")
        
    return


def stop_container(container_name=None):
    container_name = container_name or choose_container_with_prompt("running", "stop")
    if container_name:
        try:    
            dockerutils.DOCKER_CLIENT.api.kill(container_name)
            display.output_message("Container stopped!", "success")
        except Exception as err:
            display.output_message(f"Error stopping container {container_name}: {err}", "error")


def upgrade_and_restart_container(container_name=None):
    if os_platform_info == "aarch64" and nvidia.l4t_version() < "35.1":
        display.output_message(
            "Couldn't restart the container because JetPack is outdated. Please, update JetPack or contact Lumeo support.",
            "error")
        return
    
    # Setup Jetson GPIO if needed
    jetson_model_name = nvidia.get_jetson_model_name()
    if jetson_model_name:
        nvidia.setup_jetson_gpio()

    if container_name is None:
        # If no explicit `container_name` is given, prompt the user.
        container_name = choose_container_with_prompt("all", "restart")
        if container_name is None:
            return

    current_container = dockerutils.DOCKER_CLIENT.containers.get(container_name)
    if not current_container:
        display.output_message("Container not found", "error")
        return

    # Get the image used by the current container
    current_image = current_container.attrs["Config"]["Image"]

    # Check if the current image has a specific tag like ':staging' or ':development'
    if ':' in current_image:
        image_base, current_tag = current_image.split(':')
    else:
        image_base = current_image
        current_tag = None

    # Preserve the tag if it's a special one, otherwise use get_lumeod_image
    if current_tag in ['staging', 'development']:
        final_image = f"{image_base}:{current_tag}"
    else:
        final_image = get_lumeod_image()  # Fallback to get_lumeod_image if no special tag is found

    # Pull the latest version of lumeo image
    #hostutils.run_command(f"docker pull {final_image}", sudo=True)
    dockerutils.docker_download_image(final_image)

    network_mode = current_container.attrs["HostConfig"]["NetworkMode"]
    port_mapping = current_container.attrs["HostConfig"]["PortBindings"]
    container_labels = current_container.attrs["Config"]["Labels"]

    container_labels["lumeo.container_version"] = LUMEO_CONTAINER_VERSION
    container_labels["lumeo.container_type"] = final_image

    environment = {"JETSON_MODEL_NAME": jetson_model_name} if jetson_model_name else None

    dockerutils.DOCKER_CLIENT.api.remove_container(container_name, force=True)

    # We restart an existing container, so the volume exists. Let's write fresh hardware info to it.
    write_hardware_information_to_container(container_name)

    if not dockerutils.DOCKER_CLIENT.containers.run(
            final_image,  # Use the preserved staging/development or fallback to default image tag
            volumes=[container_name + ":/var/lib/lumeo"],
            restart_policy={"MaximumRetryCount": 0, "Name": "always"},
            name=container_name,
            hostname=container_name,
            network_mode=network_mode,
            ports=port_mapping,
            devices=nvidia.get_devices_for_container(),
            device_requests=device_requests(),
            runtime=nvidia.get_nvidia_docker_runtime(),
            labels=container_labels,
            detach=True,
            environment=environment,
            privileged=True,
            ulimits=[docker.types.Ulimit(name="core", soft=0, hard=0)],  # Directly set ulimits to disable core dumps
    ):
        display.output_message(f"Failed to restart container {container_name}", "error")
        return

    display.output_message(f"Container {container_name} restarted with image {final_image}!", "success")    
    return



def upgrade_v2_containers():
    # Restart would change container version from v2 to v3.
    # Restart affects
    # - `x86` v2 containers with up-to-date NVIDIA driver
    # - Jetson v2 containers with up-to-date Jetpack
    if os_platform_info == "aarch64" and nvidia.l4t_version() < "35.1":
        display.output_message("Skipping v1/v2 containers upgrade because JetPack is outdated", "error")
        return False

    v2_containers = get_lumeo_containers(version=2)

    # Skip further checks if there are no v2 containers anyway.
    if not v2_containers:
        display.output_message("No outdated containers found.", "info")
        return True

    try:
        if os_platform_info == "x86_64" and nvidia.get_dgpu_driver_version() < 515:
            display.output_message("Skipping v2 containers restart because NVIDIA dGPU driver is outdated", "error")
            return False
    except:
        display.output_message(
            "[lumeo-container-update] Can't check dGPU driver version. Please make sure GPU drivers and nvidia-smi are installed",
            "error")
        return False

    for container in v2_containers:
        upgrade_and_restart_container(container.name)

    return True


def remove_container(container_name=None):
    container_name = container_name or choose_container_with_prompt("all", "remove")
    if container_name:
        try:
            dockerutils.DOCKER_CLIENT.api.remove_container(container_name, force=True)
            dockerutils.DOCKER_CLIENT.api.remove_volume(container_name, force=True)
            display.output_message("Container removed!", "success")
        except Exception as err:
            display.output_message(f"Error removing container {container_name}: {err}", "error")


def update_containers(force_update=False):

    common.update_common_dependencies()    
    
    remove_deprovisioned_containers()

    # TODO: Stop writing hardware information every hour when all lumeod's are updated to a release newer than 1.23.x.
    write_hardware_information_for_all_containers()

    if upgrade_v2_containers() and (force_update or schedule.check_update_schedule()):
        dockerutils.watchtower_trigger_update()

    updater.install_update_gateway_updater()
    
    return


def remove_deprovisioned_containers():
    containers_list = get_lumeo_containers()

    for container in containers_list:
        try:
            if container.name != "lumeo-watchtower":
                # Read provision info from volume
                volume = dockerutils.DOCKER_CLIENT.volumes.get(container.name)
                mountpoint = volume.attrs.get("Mountpoint")

                if os.path.isfile(os.path.join(mountpoint, "lumeod.toml")):
                    toml_file = open(os.path.join(mountpoint, "lumeod.toml"), "r")
                    lumeod_config = toml.load(toml_file)

                    # Check if there's some container with deprovisioned=true label on lumeod.toml
                    if "deprovisioned" in lumeod_config:
                        if lumeod_config["deprovisioned"] is True:
                            # In that case, remove the container image & volume
                            print("[lumeo-container-update] Deprovision detected, removing the following "
                                  "Lumeo container and volume: {}".format(container.name))
                            dockerutils.DOCKER_CLIENT.api.remove_container(container.name, force=True)
                            dockerutils.DOCKER_CLIENT.api.remove_volume(container.name, force=True)
        except Exception as err:
            display.output_message("[lumeo-container-update] Exception: {}".format(err), "error")


def logs(container_name=None):
    container_name = container_name or choose_container_with_prompt("all", "print logs")
    if container_name:
        display.print_text(dockerutils.DOCKER_CLIENT.api.logs(container_name).decode("ascii"))


def shell(container_name=None):
    container_name = container_name or choose_container_with_prompt("running", "access with interactive shell")
    if container_name:
        hostutils.run_command(f"docker exec -it {container_name} bash", useOsRun=True)

def write_hardware_information_for_all_containers():
    containers_list = get_lumeo_containers()
    for container in containers_list:
        write_hardware_information_to_container(container.name)


# Write output from various hardware information tools to container volume.
# Lumeod would then parse gateway specifications from these files.

def write_hardware_information_to_container(container_name):
    try:
        display.output_message(f"Getting data directory for container {container_name}", "debug")
        volume = dockerutils.DOCKER_CLIENT.volumes.get(container_name)
        mount_point = volume.attrs['Mountpoint']
        hardware_information_dir = f"{mount_point}/hardware_information"

        display.output_message(f"Attempting to remove directory: {hardware_information_dir}", "debug")
        try:
            shutil.rmtree(hardware_information_dir)
            display.output_message(f"Removed directory: {hardware_information_dir}", "debug")
        except FileNotFoundError:
            display.output_message(f"Directory not found: {hardware_information_dir}", "debug")
            pass

        display.output_message(f"Writing hardware information to {hardware_information_dir} ...", "debug")
        write_hardware_information_to_directory(hardware_information_dir)

        display.output_message("Changing owner ...", "debug")
        lumeod_file = f"{mount_point}/lumeod.toml"
        stat_info = os.stat(lumeod_file)
        uid = stat_info.st_uid
        gid = stat_info.st_gid
        hostutils.run_command(f"chown -R {uid}:{gid} {hardware_information_dir}", sudo=True)

        display.output_message("Hardware information written.", "info")
    except IndexError:
        display.output_message("Error: The container does not have any volumes.", "error")
    except KeyError:
        display.output_message("Error: The volume does not have a source.", "error")
    except FileNotFoundError:
        display.output_message(f"Error: File not found: {lumeod_file}", "error")
    except Exception as error:
        display.output_message(f"Exception: {error}", "error")


def write_hardware_information_to_directory(directory):
    # Ensure the directory exists
    if not os.path.exists(directory):
        os.makedirs(directory)

    # Define the commands and their corresponding output files
    commands = [
        ('lshw -json', 'lshw.json'),
        ('lscpu --json', 'lscpu.json'),
        ('nvpmodel --verbose --parse', 'nvpmodel-all.txt'),
        ('nvpmodel --query', 'nvpmodel-current.txt')
    ]

    for command, file in commands:
        result = hostutils.run_command(command, silent=True)
        if result is None:
            continue

        # Write the output to the corresponding file
        with open(os.path.join(directory, file), 'w') as f:
            f.write(result)

    # Save disk info
    command = 'smartctl --json --scan'
    result = hostutils.run_command(command)
    if result is None:
        return

    disk_info = []
    scanned_disks = json.loads(result).get('devices', [])
    for disk in scanned_disks:
        disk_name = disk.get('name', '')
        command = f'smartctl --json --info {disk_name}'
        result = hostutils.run_command(command)
        if result is not None:
            disk_info.append(json.loads(result))

    # Write the output to the corresponding file
    with open(os.path.join(directory, 'smartctl.json'), 'w') as f:
        json.dump(disk_info, f, indent=4)  # Prettify JSON output



import requests
import subprocess
import sys
import os

from packaging import version

from lumeo import __version__, dist_name

def check_for_update():
    """
    Check if a new version of the package is available on PyPI.
    
    Returns:
        tuple: (bool, str) A tuple containing a boolean indicating if an update is available,
                and a string with the latest version number.
    """
    try:
        # Get the latest version from PyPI
        response = requests.get(f"https://pypi.org/pypi/{dist_name}/json")
        latest_version = version.parse(response.json()["info"]["version"])
        
        # Compare with the current version
        current_version = version.parse(__version__)
        
        if latest_version > current_version:
            return True, str(latest_version), f"{current_version} (Newer version available: {latest_version})"
        else:
            return False, str(current_version), f"{current_version} (Latest version)"
    except Exception as e:
        print(f"Error checking for updates: {e}")
        return False, __version__, f"Error checking for updates: {e}"


def self_update():
    """
    Update the package to the latest version using the same installation method.
    """

    def run_command(cmd):
        try:
            subprocess.check_call(cmd, shell=True)
        except subprocess.CalledProcessError as e:
            print(f"Error executing command: {e}")
            sys.exit(1)

    update_available, latest_version, _ = check_for_update()
    
    if not update_available:
        print(f"You are already using the latest version ({__version__}).")
        return

    print(f"Updating {dist_name} to version {latest_version}...")

    # Check if running in a virtual environment
    in_venv = sys.prefix != sys.base_prefix

    # Check if installed via pipx
    is_pipx = "PIPX_HOME" in os.environ or "PIPX_LOCAL_VENVS" in os.environ

    if is_pipx:
        run_command(f"pipx upgrade {dist_name}")
    elif in_venv:
        run_command(f"pip install --upgrade {dist_name}")
    else:
        # If not in a venv and not installed via pipx, use pip with the --user flag
        run_command(f"pip install --user --upgrade {dist_name}")

    print(f"Successfully updated {dist_name} to version {latest_version}.")


import os
import argparse
import sys

def create_package_structure(package_name, author, email, description):
    
    # Convert package name to Python package name format
    package_name = package_name.lower().replace('-', '_').replace(' ', '_')
    
    # Validate package name
    if not package_name.isidentifier():
        raise ValueError("Invalid package name. Please use only letters, numbers, and underscores, and start with a letter or underscore.")
    
    # Create main directory to hold the package and git repository
    os.makedirs(package_name)

    # Create setup.py
    setup_py = f"""
from setuptools import setup, find_packages

setup(
    name='{package_name}',
    version='0.1.0',
    packages=find_packages(),
    install_requires=[
        # List your runtime dependencies here
    ],
    extras_require={{
        'tests': [
            'pytest',
        ],
    }},
    author='{author}',
    author_email='{email}',
    description='{description}',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)
"""
    with open(os.path.join(package_name, 'setup.py'), 'w') as f:
        f.write(setup_py)

    # Create package directory
    os.makedirs(os.path.join(package_name, package_name))

    # Create __init__.py
    with open(os.path.join(package_name, package_name, '__init__.py'), 'w') as f:
        f.write("# Package initialization\n")

    # Create dummy module (similar to display.py)
    dummy_module = """
from lumeopipeline import VideoFrame
import cv2

frame_count = 0

def process_frame(frame: VideoFrame, **kwargs) -> bool:
    global frame_count
    frame_count += 1

    with frame.data() as mat:
        cv2.putText(mat, f"Frame {frame_count}", (50, 50), cv2.FONT_HERSHEY_DUPLEX, 0.7, (0, 0, 255), 1)

    return True
"""
    with open(os.path.join(package_name, package_name, 'sample_module.py'), 'w') as f:
        f.write(dummy_module)

    # Create tests directory
    os.makedirs(os.path.join(package_name, 'tests'))

    # Create __init__.py in tests directory
    with open(os.path.join(package_name, 'tests', '__init__.py'), 'w') as f:
        f.write("import pytest\n")


    readme_md = f"""
# Overview

This repository contains Lumeo Custom Functions, to be used in Lumeo's [custom function node](https://docs.lumeo.com/docs/custom-function-node).

# How to use

1. Add or replace my_module.py with your custom code. You can add other modules to the package if you want.

2. Create a new git repository inside the top level {package_name} folder. Publish the package to a git repository. 
Replace the 'url' in setup.py with the URL to your git repository.

3. Using the `Utils.install_import` method in a Lumeo custom function node.

You can install this package in a Lumeo custom function node, and call it's functions using the following code snippet. If it's a public git repo, skip the token, otherwise use a read-only [personal access token](https://github.com/settings/tokens) with the `repo` scope.

> Important: The version number must match the version number in the `setup.py` file. When you make changes to the package, 
> you must also update the package version and specify it in the `Utils.install_import` method, even if you specify a git 
> branch or tag in the `git_url` parameter.
> This is required since Lumeo will not re-download the package if the specified version already exists in the container.

```python
from lumeopipeline import VideoFrame, Utils

Utils.install_import('{package_name}', 
                     version='0.1.0',
                     git_url='https://<token>@github.com/<username>/<repo>.git')
from {package_name} import sample_module

def process_frame(frame: VideoFrame, **kwargs):
    return sample_module.process_frame(frame, **kwargs)
```

4. Alternatively to #3, you can also use the `Custom Function Repo` node in Lumeo, and configure it with the git URL to your repository.
"""

    with open(os.path.join(package_name, 'README.md'), 'w') as f:
        f.write(readme_md)

    print(f"Package '{package_name}' structure created successfully. Check the README.md file for instructions on how to use it.")


def main():
    parser = argparse.ArgumentParser(description="Create a basic Python package structure to host Lumeo Custom Functions.")
    parser.add_argument("package_name", nargs='?', help="Name of the package")
    parser.add_argument("--author", default="Your Name", help="Author's name")
    parser.add_argument("--email", default="your.email@example.com", help="Author's email")
    parser.add_argument("--description", default="A Python package that contains Lumeo Custom Functions", 
                       help="Short description of the package")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    create_package_structure(args.package_name, args.author, args.email, args.description)


if __name__ == "__main__":
    main()




import lumeo.scripts.gateway.display as display
import lumeo.scripts.gateway.hostutils as hostutils
import lumeo.scripts.gateway.nvidia as nvidia
import lumeo.scripts.gateway.docker as docker

def install_common_dependencies():
    
    display.output_message("Installing common dependencies...", status='info')
        
    hostutils.check_os()
    
    hostutils.check_disk_space()
    
    # Setup Nvidia DGPU drivers
    nvidia.check_nvidia_dgpu_driver()
    
    nvidia.check_disable_nvidia_driver_updates()
    
    # Setup Jetson specific components
    nvidia.enable_jetson_clocks()
    
    jetson_model_name = nvidia.get_jetson_model_name()
    if jetson_model_name:
        nvidia.setup_jetson_gpio()    
    nvidia.install_jetson_extras()
    
    # Setup Docker and NVIDIA toolkit
    docker.install_docker_and_nvidia_toolkit()
    
    docker.check_set_docker_data_dir()
    
    docker.set_docker_logs_configuration()
    
    docker.fix_nvidia_nvml_init_issue()
    
    # Disable X server
    hostutils.prompt_disable_x_server()
    
    # Start watchtower
    docker.start_watchtower()
    
    return

    
def update_common_dependencies():
    
    docker.fix_nvidia_nvml_init_issue()
    
    nvidia.enable_jetson_clocks()

    docker.set_docker_logs_configuration()
    
    docker.migrate_watchtower()
    
    return
    


import os
import shutil
import sys
import platform
import requests
import filecmp
import glob

import lumeo.scripts.gateway.display as display
import lumeo.scripts.gateway.hostutils as hostutils

def check_nvidia_dgpu_driver():
    """Check if the NVIDIA driver is installed and up to date."""
    MIN_NVIDIA_DRIVER_VERSION = 515
    
    arch_type = platform.machine()
    if arch_type == "x86_64":
        if not shutil.which("nvidia-smi"):
            display.output_message("Error: Can't find nvidia-smi. Lumeo container requires an NVIDIA GPU and corresponding drivers.", status='error')
            sys.exit(1)

        # Check nvidia-smi output and return error if command fails
        hostutils.run_command("nvidia-smi", check=False, error_message="nvidia-smi returned an error. Please check the output below for more information:")

        driver_version = get_dgpu_driver_version()
        if driver_version < MIN_NVIDIA_DRIVER_VERSION:
            display.output_message(f"NVIDIA dGPU driver is outdated: {driver_version}.", status='error')
            display.output_message(f"Please upgrade NVIDIA driver to {MIN_NVIDIA_DRIVER_VERSION} or higher.", status='error')
            sys.exit(1)    
            
def check_disable_nvidia_driver_updates():
    """Disable NVIDIA GPU driver updates by apt if user chooses to."""
    if display.prompt_yes_no("Disable NVIDIA GPU drivers update by apt?", "n"):
        nvidia_driver_package = hostutils.run_command("dpkg -l | grep nvidia-driver | awk '/ii/{print $2}'", check=False)
        if nvidia_driver_package:
            hostutils.run_command(f"apt-mark hold {nvidia_driver_package}", sudo=True)
            display.output_message(f"NVIDIA driver package {nvidia_driver_package} has been held from updates.", status='info')
        else:
            display.output_message("NVIDIA driver package not found. Unable to hold the package.", status='error')


def get_dgpu_driver_version():
    version_string = hostutils.run_command('nvidia-smi --query-gpu=driver_version --format=csv,noheader', check=False)
    return int(version_string.split('.')[0])



#################################################
# Jetson specific functions
#################################################

def l4t_version():
    """Get the installed Jetpack L4T version."""
    try:
        result = hostutils.run_command("apt-cache policy nvidia-l4t-core")
        for line in result.splitlines():
            if "Installed:" in line:
                return line.split()[1]
    except Exception:
        return None


def install_jetson_extras():
    arch_type = os.uname().machine
    if arch_type == "aarch64":
        l4t_ver = l4t_version()
        if l4t_ver and l4t_ver >= "36":
            hostutils.apt_install("nvidia-l4t-dla-compiler "
                        "nvidia-l4t-jetson-multimedia-api "
                        "nvidia-l4t-gstreamer "
                        "gstreamer1.0-plugins-good "
                        "gstreamer1.0-plugins-bad "
                        "gstreamer1.0-plugins-ugly "
                        "gstreamer1.0-libav",
                        update_first=True)    
    

def enable_jetson_clocks():
    """Enable Jetson clocks and set power mode for specific Jetson devices."""
    arch_type = platform.machine()    
    if arch_type == "aarch64":
        jetson_codename = os.getenv('JETSON_CODENAME')
        if jetson_codename and jetson_codename in ['galen', 'jakku']:
            jetson_target_power_mode = "0" if jetson_codename == "galen" else "8"
            hostutils.run_command(f"nvpmodel -m {jetson_target_power_mode}", sudo=True)
            hostutils.run_command("rm -rf /var/lib/nvpmodel/status", sudo=True)
            hostutils.run_command(f"sed -i -e 's/\\(< PM_CONFIG DEFAULT=\\).*\\( >\\)/< PM_CONFIG DEFAULT={jetson_target_power_mode} >/g' /etc/nvpmodel.conf", sudo=True)

        jetson_clocks_file = "/tmp/jetson_clocks.service"
        response = requests.get("https://assets.lumeo.com/lumeod/jetson_clocks.service")
        with open(jetson_clocks_file, 'wb') as file:
            file.write(response.content)
        
        hostutils.run_command(f"install -m u=rw,g=r,o=r {jetson_clocks_file} /etc/systemd/system/", sudo=True)
        hostutils.run_command("systemctl start jetson_clocks", sudo=True)
        os.remove(jetson_clocks_file)
    
    
def get_jetson_model_name():
    arch_type = platform.machine()
    if arch_type == "aarch64":
        try:
            import Jetson.GPIO as GPIO
            jetson_model_name = str(GPIO.model)                
            return jetson_model_name
        except:
            pass            
    return None


def setup_jetson_gpio():
    if get_jetson_model_name():
        # Copy udev rules for GPIO chip.
        import Jetson.GPIO as GPIO
        package_path = os.path.dirname(GPIO.__file__)
        source_path = os.path.join(package_path, '99-gpio.rules')
        destination_path = "/etc/udev/rules.d/99-gpio.rules"
        if (os.path.exists(source_path) and \
            (not os.path.exists(destination_path) or not filecmp.cmp(source_path, destination_path))):
            shutil.copyfile(source_path, destination_path)
            hostutils.run_command(["udevadm", "control", "--reload-rules"], check=True)
            hostutils.run_command(["udevadm", "trigger", "--action=add"], check=True)    
            
            


#################################################
# Common functions
#################################################

def get_nvidia_docker_runtime():
    arch_type = platform.machine()
    if arch_type == "aarch64":
        if l4t_version() < "35.1":
            return None
        else:
            # "nvidia" runtime is required starting from JetPack 5.0.
            return "nvidia"
    elif arch_type == "x86_64":
        return None
    else:
        raise RuntimeError("Unsupported platform: {}".format(arch_type))
    
def get_devices_for_container():
    gpio = glob.glob("/dev/gpiochip*") if get_jetson_model_name() else []
    all = get_usb_cameras() + gpio
    return all or None

def get_usb_cameras():
    usb_cameras_list = []
    try:
        list_of_cameras = hostutils.run_command("ls /dev/video* -R 2>/dev/null", check=False).split()
        for usb_cam in list_of_cameras:
            usb_cameras_list.append("{}:{}:rwm".format(usb_cam, usb_cam))
    except:
        pass

    return usb_cameras_list
 


import asyncio
from typing import Union
import logging
import os

from lumeo.api import LumeoApiClient
from lumeo.api.lumeo_types import JsonObject

class LumeoDeployer():
    
    def __init__(self, 
                 api_client: LumeoApiClient,
                 queuing_lock: asyncio.Lock,
                 file_uri: str = None,
                 camera_external_id: str = None, 
                 camera_id: str = None,
                 
                 stream_uri: str = None,
                 stream_id: str = None,
                 pipeline_id: str = None,
                 deployment_config: JsonObject = None,
                 gateway_id: str = None,
                 
                 tag_id: str = None,

                 deployment_prefix: str = None,
                 delete_processed_files: bool = False):
        
        self.api_client = api_client
        self.queuing_lock = queuing_lock
        self.file_uri = file_uri
        self.camera_external_id = camera_external_id
        self.camera_id = camera_id
        
        self.stream_uri = stream_uri
        self.stream_id = stream_id
        self.pipeline_id = pipeline_id
        self.deployment_config = deployment_config
        self.gateway_id = gateway_id
                
        self.tag_id = tag_id
        self.deployment_prefix = deployment_prefix or ''
        self.delete_processed_files = delete_processed_files

        
    def log_debug(self, message: str) -> None:
        logging.debug(f"[{self.file_uri or self.stream_id or self.stream_uri}] {message}")
        
    def log_info(self, message: str) -> None:
        logging.info(f"[{self.file_uri or self.stream_id or self.stream_uri}] {message}")

    def log_warning(self, message: str) -> None:
        logging.warning(f"[{self.file_uri or self.stream_id or self.stream_uri}] {message}")   

    def log_error(self, message: str) -> None:
        logging.error(f"[{self.file_uri or self.stream_id or self.stream_uri}] {message}")   
        
    async def get_file_size(self) -> int:
        # Get the size of the file
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, os.path.getsize, self.file_uri)
    
    async def delete_file(self) -> None:
        # Delete the file
        self.log_info(f"Deleting file {self.file_uri} from local folder")
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(None, os.remove, self.file_uri)

    async def get_deployment_queue_id(self) -> str:
        # Get the deployment queue for this app
        deployment_queue_id = await self.api_client.get_deployment_queue_id()
        if not deployment_queue_id:
            err = f"Deployment queue not found for app {self.api_client.application_id}"
            self.log_error(err)
            raise Exception(err)
        return deployment_queue_id                


    async def get_or_create_camera_with_external_id(self, camera_external_id) -> JsonObject:
        # Get or create a camera using the camera external identifier
        camera = await self.api_client.get_camera_with_external_id(camera_external_id)
        if not camera:
            camera = await self.api_client.create_virtual_camera(camera_external_id, camera_external_id)
        return camera


    async def get_camera_with_id(self, camera_id) -> Union[JsonObject, None]:
        # Get a camera using the camera id. If doesnt exist, return None
        return await self.api_client.get_camera_with_id(camera_id)
    

    async def upload_and_create_input_stream(self, file_path: str, camera: Union[JsonObject,None]) -> Union[JsonObject, None]:
        input_stream = None
        camera_id = camera["id"] if camera else None   
        try:
            log_msg = f"Uploading file {file_path}, creating input stream"
            if camera_id:
                log_msg += f" for camera {camera_id}"
            self.log_info(log_msg)
            
            file_name = file_path.split("/")[-1]

            # Upload the file to Lumeo cloud and delete from local dir
            file_size = await self.get_file_size()
            file_object = await self.api_client.create_file(file_name, file_size, camera_id)
            file_uploaded = await self.api_client.upload_file(file_object["data_url"], file_object["metadata_url"], file_path)
            if file_uploaded:
                await self.api_client.set_file_status(file_object["id"], "uploaded")
                # Create input stream from uploaded file
                input_stream = await self.api_client.create_lumeo_file_stream(file_object, camera_id)
                if self.tag_id:
                    await self.api_client.add_tag_to_file(file_object["id"], self.tag_id)
                    await self.api_client.add_tag_to_stream(input_stream["id"], self.tag_id)                                
        except Exception as e:
            self.log_error(f"File {file_path} does not exist. Skipping. Error: {e}")
            
        return input_stream
    
    
    async def create_file_input_stream(self, file_url: str, camera: Union[JsonObject,None]) -> Union[JsonObject, None]:
        camera_id = None
        log_msg = f"Creating file input stream for file {file_url}"
        if camera:
            log_msg += f" for camera {camera['id']}"
            camera_id = camera['id']
            
        self.log_info(log_msg)
        file_name = file_url.split("/")[-1].split("?")[0]        
        input_stream = await self.api_client.create_file_stream(file_name, file_url, camera_id)        
        if input_stream:
            if self.tag_id:
                await self.api_client.add_tag_to_stream(input_stream["id"], self.tag_id)
        return input_stream
                                   
    async def create_live_input_stream(self, live_url: str, gateway_id: Union[str,None]) -> Union[JsonObject, None]:
        self.log_info(f"Creating live input stream for url {live_url}")
        stream_name = live_url.split("/")[-1].split("?")[0]
        input_stream = await self.api_client.create_live_stream(stream_name, live_url, gateway_id)  
        if input_stream:
            if self.tag_id:
                await self.api_client.add_tag_to_stream(input_stream["id"], self.tag_id)
        return input_stream
                                                
    async def queue_universal_bridge_deployment(self, input_stream, camera):   
        self.log_info(f"Queuing universal bridge deployment with input stream {input_stream['id']}")

        # Get the camera's default pipeline. Clone a reference
        # deployment if one exists, otherwise, we queue a new deployment and
        # set it as the reference.
        
        async with self.queuing_lock: 
            camera = await self.get_camera_with_id(camera["id"])        
            default_pipeline_id = camera["reference_pipeline_ids"][0] if camera["reference_pipeline_ids"] else None
            reference_deployment_id = camera["reference_deployment_ids"][0] if camera["reference_deployment_ids"] else None

            if default_pipeline_id:
                # Get pipeline name
                pipeline = await self.api_client.get_pipeline(default_pipeline_id)
                pipeline_name = pipeline["name"] if pipeline else default_pipeline_id
                deployment_configuration = {}
                deployment_name = f"{self.deployment_prefix}{camera['name']}-{pipeline_name}"
                
                if reference_deployment_id:
                    # Clone the reference deployment's configuration
                    reference_deployment = await self.api_client.get_deployment(reference_deployment_id)
                    if reference_deployment:
                        deployment_configuration = reference_deployment["configuration"]
                        deployment_configuration["video1"]["source_id"] = input_stream["id"]
                        deployment_configuration["video1"]["source_type"] = "stream"
                else:
                    # Create a new deployment configuration
                    deployment_configuration = {"video1": {"source_type": "stream", "source_id": input_stream["id"]}}
                    deployment_name += " reference deployment"

                success, deployment_id = await self.queue_deployment(camera["id"], None, 
                                                                     default_pipeline_id, deployment_configuration, deployment_name)

                if success and deployment_id and not reference_deployment_id:
                    # Set the queued deployment as the reference deployment
                    await self.api_client.set_camera_reference_deployments(camera["id"], deployment_id)
                    
                return success, deployment_id
            
            else:
                self.log_warning(f"Camera {camera['id']} does not have a default pipeline. Cannot queue deployment.")
        
        return False, None
            
                        
    async def process_new_file_and_get_stream(self):                        
        # If the file is not a valid file, skip it.
        file_name = self.file_uri.split("/")[-1].split("?")[0]
        if file_name.split(".")[-1] not in ["mp4", "avi", "mov", "mkv", "wmv", "flv", "webm", "jpg", "jpeg", "png"]:
            self.log_warning(f"File is not a video file. Skipping : {self.file_uri}")
            return None
                        
        self.log_info(f"Processing file {file_name} with camera_external_id={self.camera_external_id} / camera_id={self.camera_id}")        
        
        # Get or create a camera using the camera external identifier
        camera = None
        if self.camera_external_id:
            camera = await self.get_or_create_camera_with_external_id(self.camera_external_id)
        elif self.camera_id:
            camera = await self.get_camera_with_id(self.camera_id)

        if camera:
            self.log_debug(f"Camera found: {camera['id']} {camera['external_id']} {camera['name']}")
        elif self.pipeline_id is None and (self.camera_external_id or self.camera_id):
            self.log_error(f"Camera with external id {self.camera_external_id} or id {self.camera_id} not found or could not be created.")
            return None

        # Get the size of the file
        input_stream = None
        if not self.file_uri.startswith("http"):
            input_stream = await self.upload_and_create_input_stream(self.file_uri, camera)
        else:
            input_stream = await self.create_file_input_stream(self.file_uri, camera)
                                                
        return input_stream, camera
    
    
    async def create_or_queue_regular_deployment(self, input_stream):
        action = "Creating" if self.gateway_id else "Queuing"
        self.log_info(f"{action} regular deployment with input stream {input_stream['id']} / pipeline_id={self.pipeline_id}")
        success = False
        deployment_id = None
        async with self.queuing_lock: 
            deployment_configuration = self.deployment_config if self.deployment_config else {}
            
            pipeline = await self.api_client.get_pipeline(self.pipeline_id)                        
            if pipeline:
                deployment_name = f"{self.deployment_prefix}{input_stream['name']}-{pipeline['name']}"
                
                if 'video1' not in deployment_configuration:
                    deployment_configuration['video1'] = {}
                deployment_configuration['video1']["source_id"] = input_stream['id']
                deployment_configuration['video1']["source_type"] = "stream"
                
                if self.gateway_id:
                    success, deployment_id = await self.create_deployment(input_stream['id'], self.pipeline_id, deployment_configuration, deployment_name)
                else:
                    success, deployment_id = await self.queue_deployment(None, input_stream['id'], self.pipeline_id, deployment_configuration, deployment_name)                
            else:
                self.log_error(f"Failed {action} deployment for file_stream_id={input_stream['id']} : Pipeline with id {self.pipeline_id} not found.")                                        
        return success, deployment_id
    
    
    async def queue_deployment(self, camera_id, stream_id, pipeline_id, deployment_configuration, deployment_name):
        success = False
        deployment_id = None
        deployment_queue_id = await self.get_deployment_queue_id()
        queued_deployment = await self.api_client.queue_deployment(
            deployment_queue_id, pipeline_id, deployment_configuration, deployment_name
        )                

        if queued_deployment:                    
            if "deployment_id" in queued_deployment:
                deployment_id = queued_deployment["deployment_id"]
                entry_id = queued_deployment["id"]
            else:
                deployment_id = queued_deployment["id"]
                entry_id = None
                        
            # Log a deployed event
            if camera_id:
                await self.api_client.create_event(
                    "api.deployment_queued", "info", f"Queued deployment for {camera_id}.", None, camera_id, None
                )
            else:
                await self.api_client.create_event(
                    "api.deployment_queued", "info", f"Queued deployment for {stream_id}.", None, None, stream_id
                )    
        
            self.log_info(f"Success queueing deployment. Entry ID: {entry_id}, Deployment ID: {deployment_id}")  
            success = True
        else:
            if camera_id:
                self.log_error(f"Failed queueing deployment for camera_id={camera_id}, input_stream_id={stream_id} / pipeline_id={pipeline_id}")
            else:
                self.log_error(f"Failed queueing deployment for stream_id={stream_id} / pipeline_id={pipeline_id}")
        
        return success, deployment_id
    
    
    async def create_deployment(self, stream_id, pipeline_id, deployment_configuration, deployment_name):
        success = False
        deployment_id = None
        deployment = await self.api_client.create_deployment(
            pipeline_id, self.gateway_id, deployment_configuration, deployment_name
        )
        if deployment:
            deployment_id = deployment["id"]
            self.log_info(f"Success creating deployment for stream_id={stream_id} / pipeline_id={pipeline_id} / gateway_id={self.gateway_id}. Deployment ID: {deployment_id}")
            success = True
        else:
            self.log_error(f"Failed creating deployment for stream_id={stream_id} / pipeline_id={pipeline_id} / gateway_id={self.gateway_id}")
        return success, deployment_id
    
    
    async def process(self):
        success = False
        deployment_id = None
        if self.file_uri:
            input_stream, camera = await self.process_new_file_and_get_stream()
            
            #Queue the deployment
            if input_stream and camera:
                success, deployment_id = await self.queue_universal_bridge_deployment(input_stream, camera)
            elif input_stream and self.pipeline_id:
                success, deployment_id = await self.create_or_queue_regular_deployment(input_stream)
                
            self.log_info(f"Processed file")
                    
            # Delete the file if it was processed successfully and delete_processed_files is True
            if success and self.delete_processed_files and not self.file_uri.startswith("http"):
                self.log_info(f"Deleting file from local folder")
                await self.delete_file()                            
            
        elif self.stream_uri and self.pipeline_id:
            input_stream = await self.create_live_input_stream(self.stream_uri, self.gateway_id)
            if input_stream:
                success, deployment_id = await self.create_or_queue_regular_deployment(input_stream)
            
        elif self.stream_id and self.pipeline_id:            
            input_stream = await self.api_client.get_stream_with_id(self.stream_id)
            if input_stream:
                success, deployment_id = await self.create_or_queue_regular_deployment(input_stream)            
            
            self.log_info(f"Processed existing stream")
            
        else:
            self.log_error("Either new file path/url and camera id/external id must be provided, or existing tag and pipeline id must be provided.")
            
        return success, deployment_id


import re
import os
import pytz

from datetime import datetime, timedelta

from tzlocal import get_localzone

import lumeo.scripts.gateway.display as display

LUMEO_UPDATE_SCHEDULE_FILE = "/opt/lumeo/update_schedule_plan.txt"

#################################################
# Native impl based Time parser and checker functions
#################################################

def parse_time(time_str):
    # Parse time with AM/PM or without
    pattern = ['%I:%M%p', '%I:%M %p'] if ':' in time_str else ['%I%p', '%I %p']
    try:
        time = datetime.strptime(time_str, pattern[0]).time()
    except:
        time = datetime.strptime(time_str, pattern[1]).time()
    return time


def parse_date(date_str):
    # Parse date with different formats
    try:
        date = datetime.strptime(date_str, '%d/%m/%Y').date()
    except ValueError:
        try:
            date = datetime.strptime(date_str, '%Y/%m/%d').date()
        except ValueError:
            raise ValueError(f"Invalid date format. Date must be in either 'd/m/Y' or 'Y/m/d' format.")
    return date


def parse_time_range(time_range_str):
    time_ranges = []

    # Define regex patterns
    date_regex = r'\d{1,2}/\d{1,2}/\d{4}|\d{4}/\d{1,2}/\d{1,2}'
    time_regex = r'\d{1,2}:\d{2}\s?[ap]m|\d{1,2}\s?[ap]m'
    day_regex = r'monday|tuesday|wednesday|thursday|friday|saturday|sunday|weekdays|weekends'

    # Find matches
    dates = re.findall(date_regex, time_range_str, re.IGNORECASE)
    times = re.findall(time_regex, time_range_str, re.IGNORECASE)
    days = re.findall(day_regex, time_range_str, re.IGNORECASE)

    # Check if explicit dates are provided
    if dates:
        raise ValueError("Explicit dates are not supported in time ranges.")

    # Convert dates and times
    dates = [parse_date(date) if date else None for date in dates]
    times = [parse_time(time) for time in times]

    # Check if there's no end time specified
    if len(times) == 1:
        times.append(datetime.strptime('2359', '%H%M').time())  # Default end time to 11:59 PM

    # Prepare the result
    if days and (days[0] == 'weekdays' or days[0] == 'weekends'):
        # Substitute weekdays with monday-friday and weekends with saturday-sunday
        substitute_days = ['monday', 'tuesday', 'wednesday', 'thursday', 'friday'] if days[0] == 'weekdays' else [
            'saturday', 'sunday']
        for day in substitute_days:
            time_range = {
                'start_date': None,
                'end_date': None,
                'start_dow': day,
                'end_dow': None,
                'start_time': times[0] if times else None,
                'end_time': times[1] if len(times) > 1 else None
            }
            validate_time_range(time_range, time_range_splitted=False)
            time_ranges.append(time_range)
    elif not (days or dates) and len(times) > 1 and times[0] > times[1]:
        # If no days or dates are specified, and the start time is after the end time,
        # then the time range spans midnight. Split it into two time ranges.

        # Check if the time range covers at least a 2-hour interval before we split it into two time ranges
        # Convert times to datetime objects
        start_time_dt = datetime.combine(datetime.min, times[0])
        end_time_dt = datetime.combine(datetime.min, times[1])

        # Adjust end time if it's on the next day
        if times[0] > times[1]:
            end_time_dt += timedelta(days=1)  # Add 1 day to end time

        # Calculate time difference in minutes
        time_diff_minutes = abs((end_time_dt - start_time_dt).total_seconds()) / 60

        if time_diff_minutes < 120:  # 2 hours
            raise ValueError("Time range must cover at least a 2-hour interval.")

        time_range = {
            'start_date': None,
            'end_date': None,
            'start_dow': None,
            'end_dow': None,
            'start_time': times[0],
            'end_time': datetime.strptime('2359', '%H%M').time()
        }
        validate_time_range(time_range, time_range_splitted=True)
        time_ranges.append(time_range)
        time_range = {
            'start_date': None,
            'end_date': None,
            'start_dow': None,
            'end_dow': None,
            'start_time': datetime.strptime('0000', '%H%M').time(),
            'end_time': times[1]
        }
        validate_time_range(time_range, time_range_splitted=True)
        time_ranges.append(time_range)
    else:
        time_range = {
            'start_date': dates[0] if dates else None,
            'end_date': dates[1] if len(dates) > 1 else None,
            'start_dow': days[0] if days else None,
            'end_dow': days[1] if len(days) > 1 else None,
            'start_time': times[0] if times else None,
            'end_time': times[1] if len(times) > 1 else None
        }
        validate_time_range(time_range, time_range_splitted=False)
        time_ranges.append(time_range)

    return time_ranges


def validate_time_range(time_range, time_range_splitted=False):
    # Check that result contains at least a start time.
    if not time_range['start_time']:
        raise ValueError("Start time must be specified.")

    # Check that result either contains days of the week or dates or neither. You can't mix DOW and dates.
    if (time_range['start_dow'] or time_range['end_dow']) and (time_range['start_date'] or time_range['end_date']):
        raise ValueError("Cannot mix days of the week with explicit dates.")

    # Check that end is not specified without start
    if (not time_range['start_dow'] and time_range['end_dow']) or (
            not time_range['start_date'] and time_range['end_date']):
        raise ValueError("Cannot specify end day/date without start.")

    # Check if start day/date/time is before the end/day/date/time when both are specified
    if time_range['start_date'] and time_range['end_date']:
        if time_range['start_date'] > time_range['end_date']:
            raise ValueError("Start date must be on or before the end date.")

    if (not time_range['start_dow'] and not time_range['start_date']) and \
            time_range['start_time'] and time_range['end_time'] and \
            time_range['start_time'] > time_range['end_time']:
        raise ValueError("Start time must be on or before the end time when dates / day of week are not specified.")

    # Check if the time range covers at least a 2-hour interval
    if not time_range_splitted and time_range.get('start_time') and time_range.get('end_time'):
        if time_range['start_dow'] and time_range['end_dow'] and time_range['start_dow'] != time_range['end_dow']:
            start_dow = datetime.strptime(time_range['start_dow'], '%A').date()
            end_dow = datetime.strptime(time_range['end_dow'], '%A').date()

            # If start and end days are different, we need to calculate the time difference considering both days
            start_time_minutes = time_range['start_time'].hour * 60 + time_range['start_time'].minute
            end_time_minutes = time_range['end_time'].hour * 60 + time_range['end_time'].minute

            # Calculate the time difference within the same day
            same_day_diff = (24 * 60) - abs(end_time_minutes - start_time_minutes)

            # Calculate the time difference across multiple days
            across_days_diff = (end_dow - start_dow).days * 24 * 60

            # Final time difference is the sum of both
            time_diff_minutes = same_day_diff + across_days_diff

            if time_diff_minutes < 120:
                raise ValueError("Time range must cover at least a 2-hour interval.")
        else:
            time_diff_minutes = abs((time_range['end_time'].hour * 60 + time_range['end_time'].minute) -
                                    (time_range['start_time'].hour * 60 + time_range['start_time'].minute))
            if time_diff_minutes < 120:  # 2 hours
                raise ValueError(f"Time range must cover at least a 2-hour interval.")

    return True


def is_datetime_in_range(time_range, now, tz, debug=False):
    # Convert day of week to date if needed
    if time_range['start_dow']:
        dow_mapping = {
            "monday": 0, "tuesday": 1, "wednesday": 2,
            "thursday": 3, "friday": 4, "saturday": 5, "sunday": 6
        }

        # Adjust dates to the nearest matching day of the week
        start_dow_offset = dow_mapping[time_range['start_dow']] - now.weekday()
        end_dow_offset = dow_mapping[time_range['end_dow']] - now.weekday() if time_range[
            'end_dow'] else start_dow_offset

        if end_dow_offset < start_dow_offset:
            if end_dow_offset == 0:
                start_dow_offset -= 7
            elif start_dow_offset == 0:
                end_dow_offset += 7
            else:
                end_dow_offset += 7

        start_date = now.date() + timedelta(days=start_dow_offset)
        end_date = now.date() + timedelta(days=end_dow_offset)

    # Use provided dates if available
    else:
        start_date = time_range['start_date'] if time_range['start_date'] else now.date()
        end_date = time_range['end_date'] if time_range['end_date'] else start_date

    # Handle time span across days
    end_time = time_range['end_time'] if time_range['end_time'] else time_range['start_time']
    start_datetime = tz.localize(datetime.combine(start_date, time_range['start_time']))
    end_datetime = tz.localize(datetime.combine(end_date, end_time))

    # Adjust for end time earlier than start time
    if (not (time_range['end_date'] or time_range['end_dow'])) and \
            time_range['start_time'] and time_range['end_time'] and \
            time_range['start_time'] > time_range['end_time']:
        end_datetime += timedelta(days=1)

    result = start_datetime <= now <= end_datetime

    if debug:
        print(f"[{result}] {time_range} : {start_datetime} - {now} - {end_datetime}\n")

    return result


def check_update_schedule():
    try:
        if not os.path.exists(LUMEO_UPDATE_SCHEDULE_FILE):
            return True  # Early return if the file doesn't exist

        with open(LUMEO_UPDATE_SCHEDULE_FILE, "r") as file:
            schedule_plan = file.read()
            if schedule_plan:
                local_timezone = pytz.timezone(datetime.now(tz=get_localzone()).strftime('%Z'))

                for time_range_str in schedule_plan.split(","):
                    parsed_time_ranges = parse_time_range(time_range_str.lower())
                    if parsed_time_ranges:
                        time_ranges = parsed_time_ranges
                        # Are we at or within the specified time range ?
                        if time_ranges:
                            time_range_result = False
                            now = datetime.now(local_timezone)

                            for time_range in time_ranges:
                                time_range_result = is_datetime_in_range(time_range, now,
                                                                         local_timezone) or time_range_result
                                if time_range_result:
                                    return True
                display.output_message("[lumeo-container-update] Current time outside update schedule. Skipping update.",
                               "info")
                return False
    except Exception as err:
        display.output_message("Update Schedule file does not exist or has wrong format. "
                       "Will proceed with update check. Exception: {}".format(err), "error")

    return True


def edit_update_schedule():
    if os.path.exists(LUMEO_UPDATE_SCHEDULE_FILE):
        # Print the current schedule
        with open(LUMEO_UPDATE_SCHEDULE_FILE, "r") as file:
            schedule = file.read().strip()  # Read file content and remove leading/trailing whitespace
            display.output_message(f"Current update schedule: {schedule}", status='info')
    else:
        display.output_message("No update schedule found.", status='info')

    while True:
        try:
            user_input = display.prompt_input("\nEnter the update schedule or write 'remove' to delete existing.\n\n"
                               "Format: [day of week | weekdays | weekends] start time - end time\n"
                               "Example: friday 10am - 6pm or weekdays 12am - 3am\n")

            if user_input.strip().lower() == "":
                # Break the loop if the input is empty
                display.output_message("Returning to the main menu, no changes were applied to the Update Schedule", status='info')
                break

            if user_input.strip().lower() == "remove":
                # Remove the schedule file if it exists
                if os.path.exists(LUMEO_UPDATE_SCHEDULE_FILE):
                    os.remove(LUMEO_UPDATE_SCHEDULE_FILE)
                    display.output_message("Update schedule removed.", status='info')
                else:
                    display.output_message("No update schedule to remove.", status='info')
                break
            else:
                # Parse the update schedule string to check for valid time ranges
                for time_range_str in user_input.split(","):
                    parse_time_range(time_range_str.lower())

                # Write the schedule line to the file
                with open(LUMEO_UPDATE_SCHEDULE_FILE, "w") as file:
                    file.write(user_input.strip())
                display.output_message("Update schedule successfully saved.", status='info')
                break
        except Exception as error:
            display.output_message(f"Invalid update schedule format, please retry! Exception: {error}", "error")


def save_update_schedule_from_env():
    # Store the contents of LUMEO_UPDATE_SCHEDULE environment variable to file
    update_schedule = os.getenv("LUMEO_UPDATE_SCHEDULE")
    if update_schedule:
        try:
            # Parse the update schedule string to check for valid time ranges
            for time_range_str in update_schedule.split(","):
                parse_time_range(time_range_str.lower())

            if not os.path.exists("/opt/lumeo"):
                os.makedirs("/opt/lumeo")

            # Write the update schedule info to file
            with open(LUMEO_UPDATE_SCHEDULE_FILE, "w") as file:
                file.write(update_schedule)
        except Exception as error:
            display.output_message(f"Error storing update schedule. Exception: {error}", "error")    


import os
import platform

import lumeo.scripts.gateway.hostutils as hostutils
import lumeo.scripts.gateway.common as common
import lumeo.scripts.gateway.display as display
import lumeo.scripts.gateway.updater as updater
import lumeo.scripts.gateway.docker as dockerutils

def get_wgm_status():
    """Get the status of the Lumeo Web Gateway Manager."""
    status = "not_installed"
    status_str = "Web Gateway Manager: "
    containers = dockerutils.DOCKER_CLIENT.containers.list(all=True, filters={"name": "lumeo_wgm"})
    if containers:
        status = containers[0].status
        container_info = containers[0].attrs        
        status_str += f"{container_info['Name']} ({container_info['Config']['Image']})"
        if status == "running":
            status_str += f" [bold green]{status.capitalize()}[/bold green]"
        else:
            status_str += f" [bold red]{status.capitalize()}[/bold red]"
    else:
        status_str += "[orange]Not Installed[/orange]"
    return status, status_str


def install_wgm():
    """Install Lumeo Web Gateway Manager."""   
    
    display.output_message("Installing Lumeo Web Gateway Manager...", status='info')    
     
    arch_type = platform.machine()
    
    wgm_status, _ = get_wgm_status()
    
    if wgm_status == 'not_installed':
        common.install_common_dependencies()
            
        # Create shared volume directory
        os.makedirs("/opt/lumeo/wgm", exist_ok=True)    
        
        # Install WGM container
        wgm_container = "lumeo/wgm-x64:latest" if arch_type == "x86_64" else "lumeo/wgm-arm:latest"

        # Remove existing lumeo_wgm container
        if hostutils.run_command("docker ps -a -q -f name=lumeo_wgm", sudo=True, check=False):
            hostutils.run_command("docker stop lumeo_wgm", sudo=True)
            hostutils.run_command("docker rm lumeo_wgm", sudo=True)
            hostutils.run_command("rm -rf /opt/lumeo/wgm", sudo=True)

        # Pull and run new container
        #run_command(f"docker pull {wgm_container}", sudo=True)
        dockerutils.docker_download_image(wgm_container)
        hostutils.run_command(f"docker run -d -v /opt/lumeo/wgm/:/lumeo_wgm/ --name lumeo_wgm --restart=always --network host {wgm_container}", sudo=True)

        # Install and start lumeo-wgm-pipe
        hostutils.run_command("install -m u=rw,g=r,o=r /opt/lumeo/wgm/scripts/lumeo-wgm-pipe.service /etc/systemd/system/", sudo=True)
        hostutils.run_command("systemctl enable --now lumeo-wgm-pipe", sudo=True)

        # Restart container
        hostutils.run_command("docker restart lumeo_wgm", sudo=True)

        # Install update cron job
        updater.install_update_gateway_updater()

        display.output_message("Lumeo Web Gateway Manager has been installed. Access by visiting https://<device-ip-address>", status='info')
    else:
        display.output_message("Lumeo Web Gateway Manager is already installed.", status='info')
    
    return


def update_wgm():
    """Update Lumeo Web Gateway Manager."""        
    
    display.output_message("Updating Lumeo Web Gateway Manager...", status='info')
    
    # Determine the appropriate container based on architecture
    arch_type = platform.machine()
    wgm_container = "lumeo/wgm-x64:latest" if arch_type == "x86_64" else "lumeo/wgm-arm:latest"

    # Get the Image ID of the currently running container
    running_image_id = hostutils.run_command("docker inspect --format='{{.Image}}' lumeo_wgm", sudo=True)

    if running_image_id:
        # Pull the latest image
        #run_command(f"docker pull {wgm_container}", sudo=True)
        dockerutils.docker_download_image(wgm_container)

        # Get the Image ID of the latest image
        latest_image_id = hostutils.run_command(f"docker inspect --format='{{{{.Id}}}}' {wgm_container}", sudo=True)

        # Compare the IDs. If they are different, stop, remove and run the new container
        if running_image_id != latest_image_id:
            hostutils.run_command("docker stop lumeo_wgm && docker rm -f lumeo_wgm", sudo=True)

            # Host network needed for bonjour broadcast
            hostutils.run_command(f"docker run -d -v /opt/lumeo/wgm/:/lumeo_wgm/ --name lumeo_wgm --restart=always --network host {wgm_container}", sudo=True)
            hostutils.run_command("systemctl restart lumeo-wgm-pipe", sudo=True)
            
            # Remove the old image
            hostutils.run_command(f"docker image rm -f {running_image_id}", sudo=True)
            display.output_message("Updated and started new WGM container successfully.", status='info')
        else:
            display.output_message("Running WGM container is up-to-date.", status='info')
            
        updater.install_update_gateway_updater()
    else:
        display.output_message("Lumeo Web Gateway Manager container not found.", status='error')
    
    return

def remove_wgm():
    """Remove Lumeo Web Gateway Manager."""
    wgm_status, _ = get_wgm_status()
    display.output_message("Removing Lumeo Web Gateway Manager...", status='info')
    
    if wgm_status != 'not_installed':
        # Stop and remove container
        hostutils.run_command("docker stop lumeo_wgm && docker rm -f lumeo_wgm", sudo=True)
        # Remove MediaMTX container if it exists
        hostutils.run_command("docker stop mediamtx && docker rm -f mediamtx", sudo=True)
        # Remove lumeo-wgm-pipe service
        hostutils.run_command("systemctl stop lumeo-wgm-pipe", sudo=True)
        hostutils.run_command("systemctl disable lumeo-wgm-pipe", sudo=True)
        hostutils.run_command("rm /etc/systemd/system/lumeo-wgm-pipe.service", sudo=True)
        hostutils.run_command("systemctl daemon-reload", sudo=True)
        # Remove shared volume directory
        hostutils.run_command("rm -rf /opt/lumeo/wgm", sudo=True)
        display.output_message("Lumeo Web Gateway Manager has been removed.", status='info')
    else:
        display.output_message("Lumeo Web Gateway Manager is not installed.", status='info')
    
    return

def reset_wgm(silent=False):
    """Reset the password for the Lumeo Web Gateway Manager."""    
    reset = True
    
    if not silent:
        display.print_header("Lumeo Web Gateway Manager Password Reset")
        display.output_message("Resetting the web password will require you to create a new device account via the web interface. "
                       "You should do so immediately, since once reset, anyone can create a new device account.")
        reset = display.prompt_yes_no("Would you like to reset the web password for this device?", "n")
    
    if reset:
        hostutils.run_command("rm -f /opt/lumeo/wgm/db.sqlite", sudo=True)
        hostutils.run_command("docker restart lumeo_wgm", sudo=True)
        display.output_message("Device account reset complete", status='info')

